import types
from typing import Callable, Awaitable, Any, List

from dico_command import Bot, Context


class PYEval:
    def __init__(self, ctx, bot, src):
        self.ctx = ctx
        self.bot = bot
        self.src = src
        self.coro = self.compile(src)

    @staticmethod
    def compile(code) -> Callable[[Bot, Context], Awaitable[Any]]:
        if code.startswith("```py") and code.endswith("```"):
            code = code.lstrip("```py\n")
            code = code.rstrip('```')
        if len(code.split('\n')) == 1 and not code.startswith("return") and not code.startswith("del") and not code.startswith("raise"):
            code = f"return {code}"
        code = '\n'.join([f'    {x}' for x in code.split('\n')])
        code = f"async def evaluate_this(bot, ctx):\n{code}"
        exec(compile(code, "<string>", "exec"))
        return eval("evaluate_this")

    async def evaluate(self) -> List[Any]:
        results = []
        coro = self.coro(self.bot, self.ctx)
        if isinstance(coro, types.AsyncGeneratorType):
            async for x in coro:
                results.append(x)
        else:
            results.append(await coro)
        return results



import os
import asyncio

import dico


def parse_bytesize(bytesize: float):
    gb = round(bytesize / (1024 * 1024 * 1024), 1)
    if gb < 1:
        mb = round(bytesize / (1024 * 1024), 1)
        if mb < 1:
            return f"{bytesize}KB"
        return f"{mb}MB"
    return f"{gb}GB"


def parse_second(time: int):
    parsed_time = ""
    day = time // (24 * 60 * 60)
    time -= day * (24 * 60 * 60)
    hour = time // (60 * 60)
    time -= hour * (60 * 60)
    minute = time // 60
    time -= minute * 60
    if day:
        parsed_time += f"{day} day{'s' if day != 1 else ''} "
    if hour:
        parsed_time += f"{hour} hours "
    if minute:
        parsed_time += f"{minute} minutes "
    parsed_time += f"{time} seconds"
    return parsed_time


async def delete_wait(bot, ctx, content=None, embed=None):
    delete_button = dico.Button(style=dico.ButtonStyles.DANGER, emoji="🗑️", custom_id="trash")
    delete_button.custom_id += str(ctx.id)
    msg = await ctx.reply(content, embed=embed, component=dico.ActionRow(delete_button))
    try:
        await bot.wait("interaction_create",
                       check=lambda inter: int(inter.author) == int(ctx.author.id) and inter.data.custom_id == delete_button.custom_id,
                       timeout=60)
        await msg.delete()
    except asyncio.TimeoutError:
        delete_button.disabled = True
        await msg.edit(component=dico.ActionRow(delete_button))


def resolve_route(route: str, extension: str = ".py", sep: str = "."):
    extension = extension if extension.startswith(".") else f".{extension}"
    was_py_file = route.endswith(extension)
    initial_route = route[:-len(extension)] if was_py_file else route
    initial_route = initial_route.replace(sep, '/')
    if "*" not in initial_route:
        if was_py_file:
            return [initial_route.replace("/", sep)]
        elif os.path.isdir(initial_route):
            files = [x for x in os.listdir(initial_route)]
            if "__init__.py" in files:
                return [initial_route.replace("/", sep)]
            files = [x for x in files if x.endswith(extension) or "__init__.py" in os.listdir(os.path.join(initial_route, x))]
            files = [x[:-len(extension)] if x.endswith(extension) else x for x in files]
            return [f"{initial_route.replace('/', sep)}.{x}" for x in files]
        elif os.path.isfile(initial_route+extension):
            return [initial_route.replace("/", sep)]
        else:
            return []
    buf = ""
    for x in initial_route.split("/"):
        if x != "*":
            buf += f"{x}/"
            continue
        if initial_route[-1] != "*":
            raise ValueError(f"`*` can only be used at the end of a route.")
        if os.path.isdir(buf):
            files = [x for x in os.listdir(buf)]
            files = [x for x in files if
                     x.endswith(extension) or (not was_py_file and "__init__.py" in os.listdir(os.path.join(buf, x)))]
            files = [x[:-len(extension)] if x.endswith(extension) else x for x in files]
            return [f"{buf.replace('/', sep)}{x}" for x in files]
        else:
            return []


def interaction_check(message_id, prefix):
    def wrap(inter: dico.Interaction):
        return inter.type.message_component and inter.data.custom_id.startswith(prefix) and inter.data.custom_id.endswith(str(message_id))
    return wrap


async def receive_interaction(ctx, prefix):
    while not ctx.client.websocket_closed:
        interaction: dico.Interaction = await ctx.client.wait("interaction_create", check=interaction_check(ctx.id, prefix), timeout=30)
        if interaction.author.id != ctx.author.id:
            resp = dico.InteractionResponse(dico.InteractionCallbackType.CHANNEL_MESSAGE_WITH_SOURCE,
                                            dico.InteractionApplicationCommandCallbackData(
                                                content="You are not authorized to use this.",
                                                flags=64
                                            ))
            ctx.client.loop.create_task(interaction.create_response(resp))
            continue
        return interaction



import asyncio
import traceback

import dico
import dico_command
from .components import DP_INFO_SELECT, DPE_INFO_SELECT
from .infos import get_front_page, get_bot_info, get_sys_info, get_full_info, get_cache_info, get_addons_info
from .pager import Pager
from .pyeval import PYEval
from .utils import delete_wait, resolve_route, receive_interaction,interaction_check


class DPAddon(dico_command.Addon, name="dp"):
    command_name = "dp"

    async def addon_check(self, ctx):
        owner_ids = await self.bot.get_owners()
        return ctx.author.id in owner_ids

    @dico_command.command(command_name)
    async def dp(self, ctx: dico_command.Context):
        info = DP_INFO_SELECT.copy()
        info.custom_id += str(ctx.id)
        msg = await ctx.reply(get_front_page(self.bot, ctx), component=dico.ActionRow(info))

        while not self.bot.websocket_closed:
            try:
                interaction: dico.Interaction = await receive_interaction(ctx, "dpinfo")
                await interaction.create_response(dico.InteractionResponse(callback_type=dico.InteractionCallbackType.DEFERRED_UPDATE_MESSAGE, data={}))
                if interaction.data.values[0] == "main":
                    await msg.edit(content=get_front_page(self.bot, ctx))
                elif interaction.data.values[0] == "bot":
                    await msg.edit(content=get_bot_info(self.bot, ctx.guild_id))
                elif interaction.data.values[0] == "sys":
                    await msg.edit(content=get_sys_info())
            except asyncio.TimeoutError:
                break
        info.disabled = True
        await msg.edit(component=dico.ActionRow(info))

    @dp.subcommand("bot")
    async def dp_bot(self, ctx: dico_command.Context):
        await ctx.reply(get_bot_info(self.bot, ctx.guild_id))

    @dp.subcommand("sys")
    async def dp_sys(self, ctx: dico_command.Context):
        await ctx.reply(get_sys_info())

    @dp.subcommand("cache")
    async def dp_cache(self, ctx: dico_command.Context):
        # await ctx.reply(get_cache_info(self.bot))
        delete_button = dico.Button(style=dico.ButtonStyles.DANGER, label="Reset", emoji="🗑️", custom_id=f"trash{ctx.id}")
        msg = await ctx.reply(get_cache_info(self.bot), components=[dico.ActionRow(delete_button)])
        try:
            interaction = await receive_interaction(ctx, "trash")
            options = [dico.SelectOption(label=x, value=x) for x in self.bot.cache.available_cache_types if x != "guild_cache"]
            select_cache = dico.SelectMenu(custom_id=f"cachesel{ctx.id}", options=[dico.SelectOption(label="All", value="all"), *options])
            resp = dico.InteractionResponse(dico.InteractionCallbackType.CHANNEL_MESSAGE_WITH_SOURCE,
                                            dico.InteractionApplicationCommandCallbackData(
                                                content="Select cache to reset",
                                                flags=64,
                                                components=[dico.ActionRow(select_cache)]
                                            ))
            await interaction.create_response(resp)
            try:
                select_resp: dico.Interaction = await receive_interaction(ctx, "cachesel")
                selected = select_resp.data.values[0]
                yes_button = dico.Button(style=dico.ButtonStyles.SUCCESS, emoji="⭕", custom_id=f"confy{msg.id}")
                no_button = dico.Button(style=dico.ButtonStyles.DANGER, emoji="❌", custom_id=f"confn{msg.id}")
                resp = dico.InteractionResponse(callback_type=dico.InteractionCallbackType.UPDATE_MESSAGE,
                                                data=dico.InteractionApplicationCommandCallbackData(
                                                    content=f"Are you sure want to reset {selected} cache?",
                                                    flags=64,
                                                    components=[dico.ActionRow(yes_button, no_button)]))
                await select_resp.create_response(resp)
                yes_button.disabled = True
                no_button.disabled = True
                try:
                    inter: dico.Interaction = await self.bot.wait("interaction_create",
                                                                  check=interaction_check(msg.id, "conf"), timeout=30)
                    if inter.data.custom_id.startswith("confy"):
                        self.bot.cache.reset(selected if selected != "all" else None)
                        resp = dico.InteractionResponse(
                            callback_type=dico.InteractionCallbackType.UPDATE_MESSAGE,
                            data=dico.InteractionApplicationCommandCallbackData(content="Successfully cleared cache.",
                                                                                flags=64,
                                                                                components=[dico.ActionRow(yes_button, no_button)])
                        )
                        await inter.create_response(resp)
                    else:
                        resp = dico.InteractionResponse(
                            callback_type=dico.InteractionCallbackType.UPDATE_MESSAGE,
                            data=dico.InteractionApplicationCommandCallbackData(content="Cancelled cache reset.",
                                                                                flags=64,
                                                                                components=[dico.ActionRow(yes_button, no_button)])
                        )
                        await inter.create_response(resp)
                except asyncio.TimeoutError:
                    await select_resp.edit_original_response(content="Cancelled cache reset.", components=[dico.ActionRow(yes_button, no_button)])
            except asyncio.TimeoutError:
                select_cache.disabled = True
                await interaction.edit_original_response(components=[dico.ActionRow(select_cache)])
        except asyncio.TimeoutError:
            pass
        delete_button.disabled = True
        await msg.edit(components=[dico.ActionRow(delete_button)])

    @dp.subcommand("addon")
    async def dp_addon(self, ctx: dico_command.Context, option: str = None, addon: str = "dp"):
        if not option:
            return await ctx.reply(get_addons_info(self.bot))
        if option not in ["load", "unload", "reload"]:
            return await ctx.reply(f"Invalid option `{option}`. Must be one of `load`, `unload`, `reload`.")
        try:
            resolved = resolve_route(addon) if addon != "dp" else ["dp"]
            if not resolved:
                return await ctx.reply("No addons to reload found.")
        except ValueError as ex:
            return await ctx.reply(str(ex))
        action = {"load": self.bot.load_module, "unload": self.bot.unload_module, "reload": self.bot.reload_module}[option]
        resp = []
        if resolved:
            for x in resolved:
                try:
                    action(x)
                    resp.append(f"✅ `{x}`")
                except Exception as e:
                    resp.append(f"❌ `{x}` - `{e}`")
        await ctx.reply("\n".join(resp))

    @dp.subcommand("py")
    async def dp_py(self, ctx: dico_command.Context, *, src: str):
        delete_button = dico.Button(style=dico.ButtonStyles.DANGER, emoji="🗑️", custom_id="trash")
        try:
            ev = PYEval(ctx, self.bot, src)
            resp = await ev.evaluate()
            str_resp = [f"Result {i}:\n{x}" for i, x in enumerate(resp, start=1)] if len(resp) > 1 else [*map(str, resp)]
            pages = []
            for x in str_resp:
                if len(x) > 2000:
                    while len(x) > 2000:
                        pages.append(x[:2000])
                        x = x[2000:]
                    pages.append(x)
                else:
                    pages.append(x)
            if len(pages) == 1:
                return await delete_wait(self.bot, ctx, content=str(pages[0]))
            pager = Pager(self.bot, ctx.channel, ctx.author, pages, reply=ctx, extra_button=delete_button, timeout=60)
            async for _ in pager.start():
                await pager.message.delete()
                break
        except Exception as ex:
            tb = ''.join(traceback.format_exception(type(ex), ex, ex.__traceback__))
            tb = ("..." + tb[-1997:]) if len(tb) > 2000 else tb
            await delete_wait(self.bot, ctx, content=tb)


class DPEAddon(dico_command.Addon, name="dpe"):
    command_name = "dpe"

    async def addon_check(self, ctx):
        owner_ids = await self.bot.get_owners()
        return ctx.author.id in owner_ids

    @dico_command.command(command_name)
    async def dpe(self, ctx: dico_command.Context):
        info = DPE_INFO_SELECT.copy()
        info.custom_id += str(ctx.id)
        msg = await ctx.reply(embed=get_front_page(self.bot, ctx, embed=True), component=dico.ActionRow(info))

        while not self.bot.websocket_closed:
            try:
                interaction: dico.Interaction = await receive_interaction(ctx.id, "dpeinfo")
                await interaction.create_response(dico.InteractionResponse(callback_type=dico.InteractionCallbackType.DEFERRED_UPDATE_MESSAGE, data={}))
                if interaction.data.values[0] == "main":
                    await msg.edit(embed=get_front_page(self.bot, ctx, embed=True))
                elif interaction.data.values[0] == "bot":
                    await msg.edit(embed=get_bot_info(self.bot, ctx.guild_id, embed=True, timestamp=ctx.timestamp))
                elif interaction.data.values[0] == "sys":
                    await msg.edit(embed=get_sys_info(embed=True, timestamp=ctx.timestamp))
                elif interaction.data.values[0] == "full":
                    await msg.edit(embed=get_full_info(self.bot, ctx))
            except asyncio.TimeoutError:
                break
        info.disabled = True
        await msg.edit(component=dico.ActionRow(info))

    @dpe.subcommand("bot")
    async def dpe_bot(self, ctx: dico_command.Context):
        await ctx.reply(embed=get_bot_info(self.bot, ctx.guild_id, embed=True, timestamp=ctx.timestamp))

    @dpe.subcommand("sys")
    async def dpe_sys(self, ctx: dico_command.Context):
        await ctx.reply(embed=get_sys_info(embed=True, timestamp=ctx.timestamp))

    @dpe.subcommand("full")
    async def dpe_full(self, ctx: dico_command.Context):
        await ctx.reply(embed=get_full_info(self.bot, ctx))



import datetime
import platform

import dico
import dico_command
import psutil
from dico.utils import rgb

from . import __version__
from .utils import parse_bytesize, parse_second

try:
    from dico_interaction import __version__ as __inter_version__
except ImportError:
    __inter_version__ = None

privileged = ["GUILD_MEMBERS", "GUILD_PRESENCES"]


def get_front_page(bot, ctx, embed: bool = False):
    py_version = platform.python_version()
    py_platform = platform.platform()
    process = psutil.Process()
    start_time = int(process.create_time())
    if embed:
        inter_ver = f"\ndico-interaction `{__inter_version__}`" if __inter_version__ else ""
        intents = [k for k, v in bot.intents.values.items() if v in bot.intents]

        embed = dico.Embed(title=f"dp {__version__} Panel",
                           description=f"dico `{dico.__version__}`\ndico-command `{dico_command.__version__}`{inter_ver}\n"
                                       f"Python `{py_version}` in `{py_platform}`",
                           color=rgb(225, 225, 225),
                           timestamp=ctx.timestamp)
        embed.set_author(name=ctx.member.nick if ctx.member else ctx.author.username, icon_url=ctx.author.avatar_url())
        embed.add_field(name="Privileged Intents",
                        value=f"{', '.join([f'`{x}`' for x in intents if x in privileged])}\n"
                              f"For more info, please run `{ctx.prefix}dpe bot` or select `Bot Info`.",
                        inline=False)
        embed.add_field(name="Process Start Time",
                        value=f"<t:{start_time}:R>. (<t:{start_time}> ~)\n"
                              f"For more info, please run `{ctx.prefix}dpe sys` or select `System Info`.")
        return embed
    else:
        inter_ver = f"dico-interaction is installed and the version is `{__inter_version__}`.\n" if __inter_version__ else ""
        start_time = int(process.create_time())
        sharded_info = f"This guild is on shard `{bot.get_shard_id(ctx.guild_id) if ctx.guild_id else 0}`.\n" if bot.shards else ""
        return f"dp `{__version__}`, with dico `{dico.__version__}` and dico-command `{dico_command.__version__}` " \
            f"on Python `{py_version}` in `{py_platform}`.\n{inter_ver}" \
            f"This bot was started <t:{start_time}:R>. (<t:{start_time}> ~)\n{sharded_info}\n" \
            f"Run `{ctx.prefix}dp bot` or select `Bot Info` for more information with this bot.\n" \
            f"Run `{ctx.prefix}dp sys` or select `System Info` for more information with hardware."


def get_sys_info(embed: bool = False, timestamp: datetime.datetime = None):
    process = psutil.Process()
    memory = psutil.virtual_memory()
    uptime_sys = (datetime.datetime.now() - datetime.datetime.fromtimestamp(psutil.boot_time())).total_seconds()
    uptime_bot = (datetime.datetime.now() - datetime.datetime.fromtimestamp(process.create_time())).total_seconds()
    if embed:
        embed = dico.Embed(title=f"System Information",
                           color=rgb(225, 225, 225),
                           timestamp=timestamp)
        embed.add_field(name="CPU Usage",
                        value=f"Total: `{psutil.cpu_percent()}`%\n"
                              f"Per thread: `{'`% | `'.join([str(x) for x in psutil.cpu_percent(percpu=True)])}`%\n"
                              f"Bot process: `{process.cpu_percent()}`%",
                        inline=False)
        embed.add_field(name="RAM Usage",
                        value=f"Total: `{memory.percent}`% ({parse_bytesize(memory.used)}/{parse_bytesize(memory.total)})\n"
                              f"Bot process: (Physical: `{round(process.memory_percent('rss'), 2)}`% | "
                              f"Virtual: `{round(process.memory_percent('vms'), 2)}`%)",
                        inline=False)
        embed.add_field(name="Uptime", value=f"System: {parse_second(round(uptime_sys))} | Bot: {parse_second(round(uptime_bot))}", inline=False)
        return embed
    else:
        return f"Current CPU load is `{psutil.cpu_percent()}`%.\n" \
               f"This bot is using `{process.cpu_percent()}`% of CPU.\n" \
               f"Current RAM usage is `{memory.percent}`%. ({parse_bytesize(memory.used)}/{parse_bytesize(memory.total)})\n" \
               f"This bot is using `{round(process.memory_percent('rss'), 2)}`% of physical memory " \
               f"and `{round(process.memory_percent('vms'), 2)}`% of virtual memory.\n" \
               f"System is online for {parse_second(round(uptime_sys))} and bot is online for {parse_second(round(uptime_bot))}."


def get_bot_info(bot, guild_id, embed: bool = False, timestamp: datetime.datetime = None):
    intents = [k for k, v in bot.intents.values.items() if v in bot.intents]
    cmds = len(bot.commands)
    addons = len(bot.addons)
    if embed:
        embed = dico.Embed(title=f"Bot Information",
                           color=rgb(225, 225, 225),
                           timestamp=timestamp)
        embed.add_field(name="Commands and Addons",
                        value=f"`{cmds}` Command{'s' if cmds > 1 else ''} | `{addons}` Addon{'s' if addons > 1 else ''}")
        embed.add_field(name="Guilds and Shards",
                        value=f"In {bot.guild_count} guilds.\n" + "Sharding is not enabled." if not bot.shard_count else
                        f"Total `{bot.shard_count}` shard{'s' if bot.shard_count > 1 else ''}. "
                        f"(This guild is on shard `{bot.get_shard_id(guild_id) if guild_id else 0}`.)",
                        inline=False)
        embed.add_field(name="Enabled Intents",
                        value=f"{', '.join([f'__`{x}`__' if x in privileged else f'`{x}`' for x in intents])}",
                        inline=False)
        return embed
    else:
        p_i = [f"`{x}` intent" for x in intents if x in privileged]
        if len(p_i) > 1:
            p_i[-1] = f"and {p_i[-1]}"
        shard_text = f" with `{bot.shard_count}` shard{'s' if bot.shard_count > 1 else ''}" if bot.shard_count else ""
        addon_text = f" and `{len(bot.addons)}` addon{'s' if len(bot.addons) > 1 else ''}" if bot.addons else ""
        privileged_text = f"For privileged intents, {(', ' if len(p_i) > 2 else ' ').join(p_i)} {'are' if len(p_i) > 1 else 'is'} enabled.\n"

        return f"This bot is in `{bot.guild_count}` guild{'s' if bot.guild_count > 1 else ''}{shard_text}.\n" \
               f"This bot has `{len(bot.commands)}` command{'s' if len(bot.commands) > 1 else ''}{addon_text}.\n\n" \
               f"Enabled intents: ```\n" \
               f"{', '.join(intents)}\n" \
               f"```{privileged_text}"


def get_full_info(bot, ctx):
    py_version = platform.python_version()
    py_platform = platform.platform()
    process = psutil.Process()
    memory = psutil.virtual_memory()
    uptime_sys = (datetime.datetime.now() - datetime.datetime.fromtimestamp(psutil.boot_time())).total_seconds()
    uptime_bot = (datetime.datetime.now() - datetime.datetime.fromtimestamp(process.create_time())).total_seconds()
    inter_ver = f" | dico-interaction `{__inter_version__}`" if __inter_version__ else ""
    intents = [k for k, v in bot.intents.values.items() if v in bot.intents]

    embed = dico.Embed(title=f"dp {__version__} Panel",
                       description=f"dico `{dico.__version__}` | dico-command `{dico_command.__version__}`{inter_ver}\n"
                                   f"Python `{py_version}` in `{py_platform}`",
                       color=rgb(225, 225, 225),
                       timestamp=ctx.id.timestamp)
    embed.set_author(name=ctx.member.nick if ctx.member else ctx.author.username, icon_url=ctx.author.avatar_url())
    embed.add_field(name="Enabled Intents",
                    value=f"{', '.join([f'__`{x}`__' if x in privileged else f'`{x}`' for x in intents])}",
                    inline=False)
    embed.add_field(name="Shards",
                    value="Sharding is not enabled." if not bot.shard_count else
                    f"Total `{bot.shard_count}` shards. (This guild is on shard `{bot.get_shard_id(ctx.guild_id) if ctx.guild_id else 0}`.)",
                    inline=False)
    embed.add_field(name="CPU Usage",
                    value=f"Total: `{psutil.cpu_percent()}`%\n"
                          f"Per thread: `{'`% | `'.join([str(x) for x in psutil.cpu_percent(percpu=True)])}`%\n"
                          f"Bot process: `{process.cpu_percent()}`%",
                    inline=False)
    embed.add_field(name="RAM Usage",
                    value=f"Total: `{memory.percent}`% ({parse_bytesize(memory.used)}/{parse_bytesize(memory.total)})\n"
                          f"Bot process: (Physical: `{round(process.memory_percent('rss'), 2)}`% | "
                          f"Virtual: `{round(process.memory_percent('vms'), 2)}`%)",
                    inline=False)
    embed.add_field(name="Uptime",
                    value=f"System: {parse_second(round(uptime_sys))} | Bot: {parse_second(round(uptime_bot))}",
                    inline=False)
    return embed


def get_cache_info(bot, embed: bool = False):
    cache = bot.cache
    cache_info_dict = {}
    for x in cache.available_cache_types:
        if x == "guild_cache":
            continue
        storage = cache.get_storage(x)
        cache_info_dict[x] = {"size": storage.size, "max_size": storage.max_size}
    if not embed:
        max_size = sorted(map(lambda n: len(n)+4, cache_info_dict.keys()))[-1]
        texts = '\n'.join([f"{f'[ {k} ]'.ljust(max_size, ' ')} - {v['size']} / {v['max_size'] if v['max_size'] > 0 else '∞'}" for k, v in cache_info_dict.items()])
        return "This bot's cache info:\n" \
               f"```css\n{texts}\n```"


def get_addons_info(bot, embed: bool = False):
    if not embed:
        texts = []
        max_size = sorted(map(lambda n: len(n)+4, [x.name for x in bot.addons]))[-1]
        for addon in bot.addons:
            addon: dico_command.Addon
            commands = f"{len(addon.commands)} command{'s' if len(addon.commands) != 1 else ''}" if addon.commands else ""
            listeners = f"{len(addon.listeners)} listener{'s' if len(addon.listeners) != 1 else ''}" if addon.listeners else ""
            interactions = f"{len(addon.interactions)} interaction{'s' if len(addon.interactions) != 1 else ''}" if addon.interactions else ""
            callbacks = f"{len(addon.callbacks)} callback{'s' if len(addon.callbacks) != 1 else ''}" if addon.callbacks else ""
            autocompletes = f"{len(addon.autocompletes)} autocomplete{'s' if len(addon.autocompletes) != 1 else ''}" if addon.autocompletes else ""
            final_texts = [x for x in (commands, listeners, interactions, callbacks, autocompletes) if x]
            if len(final_texts) > 1:
                final_texts[-1] = f"and {final_texts[-1]}"
            texts.append(f"{f'[ {addon.name} ]'.ljust(max_size, ' ')} - {(', ' if len(final_texts) > 1 else ' ').join(final_texts)}")
        joined_texts = '\n'.join(texts)
        return f"```css\n{joined_texts}\n```"
